@@ -2,7 +2,7 @@
 
   digest -- hash digest functions for R
 
-  Copyright (C) 2003 - 2014  Dirk Eddelbuettel <edd@debian.org>
+  Copyright (C) 2003 - 2016  Dirk Eddelbuettel <edd@debian.org>
 
   This file is part of digest.
 
@@ -27,6 +27,7 @@
 #include <R.h>
 #include <Rdefines.h>
 #include <Rinternals.h>
+#include <Rversion.h>
 #include <inttypes.h>
 #include "sha1.h"
 #include "sha2.h"
@@ -34,6 +35,7 @@
 #include "md5.h"
 #include "zlib.h"
 #include "xxhash.h"
+#include "pmurhash.h"
 
 unsigned long ZEXPORT digest_crc32(unsigned long crc,
                                    const unsigned char FAR *buf,
@@ -43,6 +45,7 @@ static const char *sha2_hex_digits = "0123456789abcdef";
 
 
 SEXP digest(SEXP Txt, SEXP Algo, SEXP Length, SEXP Skip, SEXP Leave_raw, SEXP Seed) {
+    size_t BUF_SIZE = 1024;
     FILE *fp=0;
     char *txt;
     int algo = INTEGER_VALUE(Algo);
@@ -50,18 +53,22 @@ SEXP digest(SEXP Txt, SEXP Algo, SEXP Length, SEXP Skip, SEXP Leave_raw, SEXP Se
     int skip = INTEGER_VALUE(Skip);
     int seed = INTEGER_VALUE(Seed);
     int leaveRaw = INTEGER_VALUE(Leave_raw);
-    SEXP result = NULL;
+    SEXP result = R_NilValue;
     char output[128+1], *outputp = output;    /* 33 for md5, 41 for sha1, 65 for sha256, 128 for sha512; plus trailing NULL */
-    int nChar;
+    uint32_t nChar;
     int output_length = -1;
     if (IS_RAW(Txt)) { /* Txt is either RAW */
         txt = (char*) RAW(Txt);
+#if defined(R_VERSION) && R_VERSION >= R_Version(3,0,0)        
+        nChar = XLENGTH(Txt);
+#else        
         nChar = LENGTH(Txt);
+#endif
     } else { /* or a string */
         txt = (char*) STRING_VALUE(Txt);
         nChar = strlen(txt);
     }
-    if (skip>0) {
+    if (skip > 0 && algo < 100) {
         if (skip>=nChar) {
             nChar=0;
         } else {
@@ -170,19 +177,27 @@ SEXP digest(SEXP Txt, SEXP Algo, SEXP Length, SEXP Skip, SEXP Leave_raw, SEXP Se
         XXH64_reset(&state, seed);
         XXH64_update(&state, (uint8 *) txt, nChar);
         unsigned long long val =  XXH64_digest(&state);
+#ifdef WIN32
         sprintf(output, "%016" PRIx64, val);
+#else 
+        sprintf(output, "%016llx", val);
+#endif
+        break;
+    }
+    case 8: {     /* MurmurHash3 32 */
+        unsigned int val = PMurHash32(seed, txt, nChar);
+        sprintf(output, "%08x", val);
         break;
     }
     case 101: {     /* md5 file case */
         int j;
         md5_context ctx;
         output_length = 16;
-        unsigned char buf[1024];
+        unsigned char buf[BUF_SIZE];
         unsigned char md5sum[16];
 
         if (!(fp = fopen(txt,"rb"))) {
-            error("Cannot open input file: %s", txt);
-            return(NULL);
+            error("Cannot open input file: %s", txt); /* already covered at R level too */
         }
         if (skip > 0) fseek(fp, skip, SEEK_SET);
         md5_starts( &ctx );
@@ -209,12 +224,11 @@ SEXP digest(SEXP Txt, SEXP Algo, SEXP Length, SEXP Skip, SEXP Leave_raw, SEXP Se
         int j;
         sha1_context ctx;
         output_length = 20;
-        unsigned char buf[1024];
+        unsigned char buf[BUF_SIZE];
         unsigned char sha1sum[20];
 
         if (!(fp = fopen(txt,"rb"))) {
-            error("Cannot open input file: %s", txt);
-            return(NULL);
+            error("Cannot open input file: %s", txt); /* already covered at R level too */
         }
         if (skip > 0) fseek(fp, skip, SEEK_SET);
         sha1_starts ( &ctx );
@@ -237,12 +251,11 @@ SEXP digest(SEXP Txt, SEXP Algo, SEXP Length, SEXP Skip, SEXP Leave_raw, SEXP Se
         break;
     }
     case 103: {     /* crc32 file case */
-        unsigned char buf[1024];
+        unsigned char buf[BUF_SIZE];
         unsigned long val;
 
         if (!(fp = fopen(txt,"rb"))) {
-            error("Cannot open input file: %s", txt);
-            return(NULL);
+            error("Cannot open input file: %s", txt); /* already covered at R level too */
         }
         if (skip > 0) fseek(fp, skip, SEEK_SET);
         val  = digest_crc32(0L, 0, 0);
@@ -264,12 +277,11 @@ SEXP digest(SEXP Txt, SEXP Algo, SEXP Length, SEXP Skip, SEXP Leave_raw, SEXP Se
         int j;
         sha256_context ctx;
         output_length = 32;
-        unsigned char buf[1024];
+        unsigned char buf[BUF_SIZE];
         unsigned char sha256sum[32];
 
         if (!(fp = fopen(txt,"rb"))) {
-            error("Cannot open input file: %s", txt);
-            return(NULL);
+            error("Cannot open input file: %s", txt); /* already covered at R level too */
         }
         if (skip > 0) fseek(fp, skip, SEEK_SET);
         sha256_starts ( &ctx );
@@ -297,11 +309,10 @@ SEXP digest(SEXP Txt, SEXP Algo, SEXP Length, SEXP Skip, SEXP Leave_raw, SEXP Se
         output_length = SHA512_DIGEST_LENGTH;
         uint8_t sha512sum[output_length], *d = sha512sum;
 
-        unsigned char buf[1024];
+        unsigned char buf[BUF_SIZE];
 
         if (!(fp = fopen(txt,"rb"))) {
-            error("Cannot open input file: %s", txt);
-            return(NULL);
+            error("Cannot open input file: %s", txt); /* already covered at R level too */
         }
         if (skip > 0) fseek(fp, skip, SEEK_SET);
         SHA512_Init(&ctx);
@@ -335,12 +346,11 @@ SEXP digest(SEXP Txt, SEXP Algo, SEXP Length, SEXP Skip, SEXP Leave_raw, SEXP Se
         break;
     }
     case 106: {     /* xxhash32 */
-        unsigned char buf[1024];
+        unsigned char buf[BUF_SIZE];
         XXH32_state_t state;
 
         if (!(fp = fopen(txt,"rb"))) {
-            error("Cannot open input file: %s", txt);
-            return(NULL);
+            error("Cannot open input file: %s", txt); /* already covered at R level too */
         }
         if (skip > 0) fseek(fp, skip, SEEK_SET);
         XXH32_reset(&state, seed);
@@ -361,12 +371,11 @@ SEXP digest(SEXP Txt, SEXP Algo, SEXP Length, SEXP Skip, SEXP Leave_raw, SEXP Se
         break;
     }
     case 107: {     /* xxhash64 */
-        unsigned char buf[1024];
+        unsigned char buf[BUF_SIZE];
         XXH64_state_t state;
 
         if (!(fp = fopen(txt,"rb"))) {
-            error("Cannot open input file: %s", txt);
-            return(NULL);
+            error("Cannot open input file: %s", txt); /* already covered at R level too */
         }
         if (skip > 0) fseek(fp, skip, SEEK_SET);
         XXH64_reset(&state, seed);
@@ -383,12 +392,46 @@ SEXP digest(SEXP Txt, SEXP Algo, SEXP Length, SEXP Skip, SEXP Leave_raw, SEXP Se
         fclose(fp);
         unsigned long long val =  XXH64_digest(&state);
 
+#ifdef WIN32
         sprintf(output, "%016" PRIx64, val);
+#else
+        sprintf(output, "%016llx", val);
+#endif
+        break;
+    }
+    case 108: {     /* murmur32 */
+        unsigned int h1=seed, carry=0;
+        unsigned char buf[BUF_SIZE];
+        XXH64_state_t state;
+        size_t total_length = 0;
+
+        if (!(fp = fopen(txt,"rb"))) {
+            error("Cannot open input file: %s", txt); /* already covered at R level too */
+        }
+        if (skip > 0) fseek(fp, skip, SEEK_SET);
+        XXH64_reset(&state, seed);
+        if (length>=0) {
+            while( ( nChar = fread( buf, 1, sizeof( buf ), fp ) ) > 0
+                   && length>0) {
+                if (nChar>length) nChar=length;
+                PMurHash32_Process(&h1, &carry, buf, nChar);
+                length -= nChar;
+                total_length += nChar;
+            }
+        } else {
+            while( ( nChar = fread( buf, 1, sizeof( buf ), fp ) ) > 0) {
+                PMurHash32_Process(&h1, &carry, buf, nChar);
+                total_length += nChar;
+            }
+        }
+        fclose(fp);
+        unsigned int val = PMurHash32_Result(h1, carry, total_length);
+
+        sprintf(output, "%08x", val);
         break;
     }
     default: {
-        error("Unsupported algorithm code");
-        return(NULL);
+        error("Unsupported algorithm code"); /* should not be reached due to test in R */
     }
     } /* end switch */
 
@@ -401,5 +444,5 @@ SEXP digest(SEXP Txt, SEXP Algo, SEXP Length, SEXP Skip, SEXP Leave_raw, SEXP Se
     }
     UNPROTECT(1);
 
-    return result;
+  return result;
 }
